# A collection of tagged document views

### CLASSDEF

setClass("TaggedCorpusView",
  representation("TaggedView", "CorpusView",
    tokenizedView = "TokenizedCorpusView",
    uniqueTags = reftype("character"),
    uniqueTokenTags = reftype("character"),
    tagFreqTable = reftype("table"),
    tokenTagFreqTable = reftype("table"),
    documentTagMatrix = reftype("TermDocMatrix"),
    documentTokenTagMatrix = reftype("TermDocMatrix")))

### GENERICS

setGeneric("documentTaggedTokenMatrix",
  function (object, weight) standardGeneric("documentTaggedTokenMatrix"))
setGeneric("documentTagMatrix",
  function (object, weight) standardGeneric("documentTagMatrix"))

### CONSTRUCTION

TaggedCorpusView <- function (tokcorp, tagger) {
  documentViews <- new("NULLlist")
  uniqueTags <- new("NULLcharacter")
  uniqueTokenTags <- new("NULLcharacter")
  tagFreqTable <- new("NULLtable")
  tokenTagFreqTable <- new("NULLtable")
  documentTagMatrix <- new("NULLTermDocMatrix")
  documentTokenTagMatrix <- new("NULLTermDocMatrix")

  res <- new("TaggedCorpusView", corpus=corpus(tokcorp),
      documentViews=documentViews, tagger=tagger,
      tokenizedView=tokcorp, uniqueTags=uniqueTags, 
      uniqueTokenTags=uniqueTokenTags, 
      tagFreqTable=tagFreqTable,
      tokenTagFreqTable=tokenTagFreqTable,
      documentTagMatrix=documentTagMatrix,
      documentTokenTagMatrix=documentTokenTagMatrix)

  registerView(corpus(tokcorp), res)
  return(res)
}

### CALLBACK method

# Set all values to null for future lazy reloading.
setMethod("updateView", 
          signature("TaggedCorpusView", "ViewableObject"),
  function (view, viewable, ...) {
# NOTE: don't want to null out documentViews here
    ref(view, "uniqueTags") <- new("NULLcharacter")
    ref(view, "uniqueTokenTags") <- new("NULLcharacter")
    ref(view, "tagFreqTable") <- new("NULLtable")
    ref(view, "tokenTagFreqTable") <- new("NULLtable")
    ref(view, "documentTagMatrix") <- new("NULLTermDocMatrix")
    ref(view, "documentTokenTagMatrix") <- new("NULLTermDocMatrix")
  }
)
 
### ACCESSORS

# Return the tagged views of the underlying docs
setMethod("documentViews", "TaggedCorpusView",
  function (object)
    if (is(ref(object, "documentViews"), "NULLlist")) {
      tokdvs <- documentViews(tokenizedView(object))
      tagdvs <- lapply(tokdvs, tag, tagger(object))
      (ref(object, "documentViews") <- tagdvs)
    } else
      ref(object, "documentViews"))
      
# Return unique "token/tag" pairs from the corpus
setMethod("unique", "TaggedCorpusView",
  function (x, incomparables="missing")
    if (is(ref(x, "uniqueTokenTags"), "NULLcharacter"))
      (ref(x, "uniqueTokenTags") <-
        unique(unlist(lapply(documentViews(x), unique))))
    else
      ref(x, "uniqueTokenTags"))

# Return unique tags
setMethod("uniqueTags", "TaggedCorpusView",
  function (object)
    if (is(ref(object, "uniqueTags"), "NULLcharacter"))
      (ref(object, "uniqueTags") <-
        unique(unlist(lapply(documentViews(object), uniqueTags))))
    else
      ref(object, "uniqueTags"))

# Return "token/tag" frequency table, across all docs
setMethod("freqTable", "TaggedCorpusView",
  function (object, ...)
    if (is(ref(object, "tokenTagFreqTable"), "NULLtable"))
      (ref(object, "tokenTagFreqTable") <-
        table(unlist(lapply(documentViews(object), taggedTokens),
                     use.names=F)))
    else
      ref(object, "tokenTagFreqTable"))

# Return tag frequency table, across all docs
setMethod("tagFreqTable", "TaggedCorpusView",
  function (object)
   if (is(ref(object, "tagFreqTable"), "NULLtable"))
      (ref(object, "tagFreqTable") <-
        table(unlist(lapply(documentViews(object), tags),
                     use.names=F)))
    else
      ref(object, "tagFreqTable"))

# Returns TRUE if object contains the given "token/tag" pair
# XXX can probably move this to CorpusView.R
setMethod("has", signature(object="TaggedCorpusView",
                           value="character"),
  function (object, value, ...) return (value %in% unique(object)))

setMethod("hasTag", signature(object="TaggedCorpusView",
                              value="character"),
  function (object, value) return (value %in% uniqueTags(object)))

# Return the total frequency of the given "token/tag" pair
setMethod("freq", signature(object="TaggedCorpusView", 
                            value="character"),
  function (object, value, ...)
    if (has(object, value))
      freqTable(object)[[value]]
    else
      return (0))

# Return the total frequency of the given tag
setMethod("tagFreq", signature(object="TaggedCorpusView",
                               value="character"),
  function(object, value)
    if (hasTag(object, value))
      tagFreqTable(object)[[value]]
    else
      return (0))

# Return the document-token/tag matrix, optionally weighted
setMethod("documentTaggedTokenMatrix", 
          signature(object="TaggedCorpusView",
                    weight="missing"),
  function (object, weight=weightTf)
    if (is(ref(object, "documentTokenTagMatrix"), "NULLTermDocMatrix"))
      (ref(object, "documentTokenTagMatrix") <-
        makeDocumentTokenMatrix(lapply(documentViews(object), freqTable),
                                unique(object),
                                names(documentViews(object))))
    else
      ref(object, "documentTokenTagMatrix"))
 
setMethod("documentTaggedTokenMatrix", 
          signature(object="TaggedCorpusView",
                    weight="WeightFunction"),
  function (object, weight=weightTf)
    weightTDM(documentTaggedTokenMatrix(object), weight))

# Return the document-tag matrix, optionally weighted
setMethod("documentTagMatrix", 
          signature(object="TaggedCorpusView",
                    weight="missing"),
  function (object, weight=weightTf)
    if (is(ref(object, "documentTagMatrix"), "NULLTermDocMatrix"))
      (ref(object, "documentTagMatrix") <-
        makeDocumentTokenMatrix(lapply(documentViews(object), tagFreqTable),
                                uniqueTags(object),
                                names(documentViews(object))))
    else
      ref(object, "documentTagMatrix"))
 
setMethod("documentTagMatrix", 
          signature(object="TaggedCorpusView",
                    weight="WeightFunction"),
  function (object, weight=weightTf)
    weightTDM(documentTagMatrix(object), weight))

### UTILITIES

setMethod("show", "TaggedCorpusView",
  function (object) {
    nUniqueTok <- length(unique(tokenizedView(object)))
    nUniqueTag <- length(uniqueTags(object))
    nUniqueTT <- length(unique(object))
    nTotalTT <- sum(documentTaggedTokenMatrix(object)@Data)

    str <- paste("A tagged corpus view of with", nTotalTT,
                 "total tagged tokens,", nUniqueTT, 
                 "unique tagged tokens,", nUniqueTok, 
                 "unique tokens, and", nUniqueTag, "unique tags\n")
    cat(str)
  }
)
