# A view of a document as a sequence of token-tag pairs

### CLASSDEF

setClass("TaggedDocumentView",
  representation("TaggedView", "DocumentView",
    tokenizedView = "TokenizedDocumentView",
    tags = reftype("character"),
    tagFreqTable = reftype("table"),
    uniqueTags = reftype("character"),
    tokenTagFreqTable = reftype("table"),
    uniqueTokenTags = reftype("character")))

### GENERICS

setGeneric("tags", function (object) standardGeneric("tags"))
setGeneric("taggedTokens", 
  function (object) standardGeneric("taggedTokens"))

### CONSTRUCTION

TaggedDocumentView <- function (tokdoc, tagger) {
  tags <- new("NULLcharacter")
  tagFreqTable <- new("NULLtable")
  uniqueTags <- new("NULLcharacter")
  tokenTagFreqTable <- new("NULLtable")
  uniqueTokenTags <- new ("NULLcharacter")

  res <- new("TaggedDocumentView", corpus=corpus(tokdoc), 
      index=index(tokdoc),
      tokenizedView=tokdoc, tagger=tagger, tags=tags, 
      tagFreqTable=tagFreqTable, uniqueTags=uniqueTags,
      tokenTagFreqTable=tokenTagFreqTable, uniqueTokenTags=uniqueTokenTags)

  registerView(corpus(tokdoc), res)
  return(res)
}

### CALLBACK method

# Set all values to null for future lazy reloading.
setMethod("updateView", 
          signature("TaggedDocumentView", "ViewableObject"),
  function (view, viewable, ...) {
    ref(view, "tags") <- new("NULLcharacter")
    ref(view, "tagFreqTable") <- new("NULLtable")
    ref(view, "uniqueTags") <- new("NULLcharacter")
    ref(view, "tokenTagFreqTable") <- new("NULLtable")
    ref(view, "uniqueTokenTags") <- new ("NULLcharacter")
  }
)
 
### ACCESSORS

# Returns the tags of this view
setMethod("tags", "TaggedDocumentView",
  function (object)
    if (is(ref(object, "tags"), "NULLcharacter"))
      (ref(object, "tags") <-
        tag(tokens(tokenizedView(object)), tagger(object)))
    else
      ref(object, "tags"))

# Return token-tag pairs in "token/tag" format
setMethod("taggedTokens", "TaggedDocumentView",
  function (object)
    paste(tokens(tokenizedView(object)), tags(object), sep="/"))

# Return a frequency table of "token/tag" pairs.
setMethod("freqTable", "TaggedDocumentView",
  function (object)
    if (is(ref(object, "tokenTagFreqTable"), "NULLtable"))
      (ref(object, "tokenTagFreqTable") <-
        table(paste(tokens(object@tokenizedView), tags(object),
                    sep="/")))
    else
      ref(object, "tokenTagFreqTable"))

# Return a frequency table of tags
setMethod("tagFreqTable", "TaggedDocumentView",
  function (object)
     if (is(ref(object, "tagFreqTable"), "NULLtable"))
      (ref(object, "tagFreqTable") <- table(tags(object)))
    else
      ref(object, "tagFreqTable"))

# Return unique "token/tag" pairs
setMethod("unique", "TaggedDocumentView",
  function (x, incomparables="missing")
    if (is(ref(x, "uniqueTokenTags"), "NULLcharacter"))
      (ref(x, "uniqueTokenTags") <- row.names(freqTable(x)))
    else
      ref(x, "uniqueTokenTags"))

# Return unique tags
setMethod("uniqueTags", "TaggedDocumentView",
  function (object)
    if (is(ref(object, "uniqueTags"), "NULLcharacter"))
      (ref(object, "uniqueTags") <- row.names(tagFreqTable(object)))
    else
      ref(object, "uniqueTags"))

# See if object contains "token/tag" pair
setMethod("has", signature(object="TaggedDocumentView",
                           value="character"),
  function (object, value, ...)
      return (value %in% taggedTokens(object)))

# See if object contain tag
setMethod("hasTag", signature(object="TaggedDocumentView",
                              value="character"),
  function (object, value) return (value %in% tags(object)))

# Return the frequency of the given "token/tag" pair
setMethod("freq", signature(object="TaggedDocumentView",
                            value="character"),
  function (object, value, ...)
    if (has(object, value))
      return (freqTable(object)[[value]])
    else
      return (0))

# Returns the frequency of the given tag
setMethod("tagFreq", signature(object="TaggedDocumentView",
                               value="character"),
  function (object, value)
    if (hasTag(object, value))
      return (tagFreqTable(object)[[value]])
    else
      return (0))

# Return tokens/tags (or pairs) at given indices.  x[i, 1] returns the
# token at index i, x[i, 2] returns the tag at index i, and x[i]
# returns the token-tag pair at index i.
setMethod("[", signature(x="TaggedDocumentView"),
  function (x, i, j, drop)
    if (missing(j)) c(tokens(x@tokenizedView)[i], tags(x)[i])
    else if (j == 1) tokens(x@tokenizedView)[i]
    else if (j == 2) tags(x)[i]
    else stop("Second argument to extract must be 1 (token) or 2 (tag)."))


### UTILITIES
setMethod("show", "TaggedDocumentView",
  function (object) {
    name <- ID(document(object))
    nTTs <- length(taggedTokens(object))
    nUniqueTokens <- length(unique(tokenizedView(object)))
    nUniqueTags <- length(uniqueTags(object))
    nUniqueTTs <- length(unique(object))

    str <- paste("A tagged document view of", name, "with", nTTs, 
                 "total tagged tokens,", nUniqueTTs,
                 "unique tagged tokens,", nUniqueTokens, 
                 "unique tokens, and", nUniqueTags, "unique tags\n")
    cat(str)
  }
)
