# TokenFilter - a base class for token filters

### CLASS DEF

# see TokenFillteredView.R

### GENERICS

setGeneric("filterTokens",
  function(object, filter, tokenizer, index) standardGeneric("filterTokens"))

# All derived classes should implement this!
setGeneric("mergeFilters",
  function (f1, f2) standardGeneric("mergeFilters"))

### METHODS

# The method signature that all extenders must implement
setMethod("filterTokens", signature(object="character", filter="TokenFilter",
                              tokenizer="missing", index="missing"),
  function (object, filter, tokenizer, index)
    stop("filter must be implemented by derived class."))

# Calls the above method when a tokenizer is provided, after
# tokenizing
setMethod("filterTokens", signature(object="character", filter="TokenFilter",
                              tokenizer="Tokenizer", index="missing"),
  function (object, filter, tokenizer, index)
    filterTokens(tokenize(object, tokenizer), filter))

# We've covered characater objects, but one can also filter
# PCorpus, TokenizedDocument, and TokenizedCorpus objects.  In the
# latter two cases the view provides tokenizer information and it
# makes no sense to pass a tokenizer along.

# We also provide specializations that merge filters for chained
# TokenFilteredViews.

# Generate a TokenFilteredCorpusView from a TokenizedCorpusView
setMethod("filterTokens", signature(object="TokenizedCorpusView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="missing"),
  function (object, filter, tokenizer, index)
    TokenFilteredCorpusView(corpus(object), filter,
                            RSNL:::tokenizer(object)))

# Specialization does filter merging
setMethod("filterTokens", signature(object="TokenFilteredCorpusView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="missing"),
  function (object, filter, tokenizer, index)
    TokenFilteredCorpusView(corpus(object),
                            mergeFilters(tokenFilter(object), filter),
                            RSNL:::tokenizer(object)))

# Generate a TokenFilteredDocumentView from a TokenizedCorpusView + index
setMethod("filterTokens", signature(object="TokenizedCorpusView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="numeric"),
  function (object, filter, tokenizer, index)
    TokenFilteredDocumentView(corpus(object), index, filter,
                              RSNL:::tokenizer(object)))

# Specialization does filter merging
setMethod("filterTokens", signature(object="TokenFilteredCorpusView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="numeric"),
  function (object, filter, tokenizer, index)
    TokenFilteredDocumentView(corpus(object), index,
                              mergeFilters(tokenFilter(object), filter),
                              RSNL:::tokenizer(object)))

# Generate a TokenFilteredCorpus from a PCorpus.
setMethod("filterTokens", signature(object="PCorpus", 
                              filter="TokenFilter",
                              index="missing"),
  function(object, filter, tokenizer, index) 
    if (missing(tokenizer))
      filterTokens(tokenize(object), filter)
    else
      TokenFilteredCorpusView(object, filter, tokenizer))

# Generate a TokenFilteredDocumentView from a PCorpus + index.
setMethod("filterTokens", signature(object="PCorpus",
                              filter="TokenFilter",
                              index="numeric"),
  function (object, filter, tokenizer, index)
    if (missing(tokenizer))
      filterTokens(tokenize(object), filter, index=index)
    else
      TokenFilteredDocumentView(object, index, filter, tokenizer))

# Generate a TokenFilteredDocumentView from a TokenizedDocumentView.
setMethod("filterTokens", signature(object="TokenizedDocumentView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="missing"),
  function (object, filter, tokenizer, index)
    TokenFilteredDocumentView(corpus(object), RSNL:::index(object), filter,
                              RSNL:::tokenizer(object)))

# specialization does merging
setMethod("filterTokens", signature(object="TokenFilteredDocumentView",
                              filter="TokenFilter",
                              tokenizer="missing",
                              index="missing"),
  function (object, filter, tokenizer, index)
    TokenFilteredDocumentView(corpus(object), RSNL:::index(object),
                              mergeFilters(tokenFilter(object), filter),
                              RSNL:::tokenizer(object)))
