# A view of a CorpusRef that has been tokenized in some way.  Provides
# the user with easy access to a TermDocMatrix, unique tokens, and the
# like.

### CLASS DEF

setClass("TokenizedCorpusView",
    representation(
        "CorpusView",
        "TokenizedView",
        tokenFreqTable          =   reftype("table"),
        uniqueTokens            =   reftype("character"),
        documentTokenMatrix     =   reftype("TermDocMatrix")
    )
)

### VALIDITY

# XXX TODO

### Various functions

# Make document-token matrix
makeDocumentTokenMatrix <- function(termfreqlist, uterms, docnames) {
  i <- lapply(termfreqlist, function (x) match(names(x), uterms)[x>0])
  names(i) <- NULL
  p <- cumsum(sapply(i, length))
  i <- unlist(i) - 1L
  x <- as.numeric(unlist(termfreqlist, use.names=FALSE))
  rm(termfreqlist)
  mat <- new("dgCMatrix", p = c(0L, p), i = i, x = x[x>0],
               Dim = c(length(uterms), length(p)))
  mat <- t(mat)
  mat@Dimnames <- list(Docs = docnames, Terms = uterms)
  new("TermDocMatrix", Data=mat, Weighting = weightTf@Name)
}

### GENERICS

setGeneric("documentTokenMatrix",
  function (object, weight, ...) standardGeneric("documentTokenMatrix"))

### CONSTRUCTION

# Constructs a new TokenizedCorpusView from given corpus and tokenizer.
# This constructor is not intended to be part of the external api and
# users will generate TokenizedCorpusViews with the tokenize method.
# Always use this function to generate TCVs within the library.
TokenizedCorpusView <- function(corpus, tokenizer) {
  documentViews <- new("NULLlist")
  tokenFreqTable <- new("NULLtable")
  uniqueTokens <- new("NULLcharacter")
  documentTokenMatrix <- new("NULLTermDocMatrix")

  res<- new("TokenizedCorpusView", corpus=corpus, tokenizer=tokenizer,
            documentViews=documentViews,
            tokenFreqTable=tokenFreqTable,
            uniqueTokens = uniqueTokens,
            documentTokenMatrix=documentTokenMatrix)
  registerView(corpus, res)
  res
}

### CALLBACK method

# XXX this is a very imprecise instrument.  Should optimize by
# allowing minimal updating where possible.
setMethod("updateView",
          signature("TokenizedCorpusView", "ViewableObject"),
  function (view, viewable, ...) {
# NOTE: we don't want to null out documentViews here because they
# update dynamically themselves.
    ref(view, "tokenFreqTable") <- new("NULLtable")
    ref(view, "uniqueTokens") <- new("NULLcharacter")
    ref(view, "documentTokenMatrix") <- new("NULLTermDocMatrix")
  }
)

### ACCESSORS / MODIFIERS (lazy loading)

# Returns the documentViews of the underlying documents
setMethod("documentViews", "TokenizedCorpusView", 
  function (object) {
    if (is(ref(object, "documentViews"), "NULLlist")) {
      tdvs <- lapply(1:length(corpus(object)), 
                     function (x) tokenize(corpus(object),
                      tokenizer(object), x))
      names(tdvs) <- sapply(corpus(object), ID)
      (ref(object, "documentViews") <- tdvs)
    } else
      ref(object, "documentViews")
  }
)

# Returns the unique tokens in this view
setMethod("unique", "TokenizedCorpusView",
  function (x, incomparables="missing")
    if (is(ref(x, "uniqueTokens"), "NULLcharacter"))
      (ref(x, "uniqueTokens") <- 
        unique(unlist(lapply(documentViews(x), unique))))
    else
      ref(x, "uniqueTokens"))

# Returns a frequency table of tokens in the corpus, across all docs
setMethod("freqTable", "TokenizedCorpusView",
  function (object, ...)
    if (is(ref(object, "tokenFreqTable"), "NULLtable"))
      (ref(object, "tokenFreqTable") <-
        table(unlist(lapply(documentViews(object), tokens),
                     use.names=F)))
    else
      ref(object, "tokenFreqTable"))

# Returns TRUE if this contains the given token
setMethod("has", signature(object="TokenizedCorpusView", value="character"),
  function(object, value, ...) return(value %in% unique(object)))

# Returns the total frequency of the given token within the corpus
setMethod("freq", signature(object="TokenizedCorpusView",
                            value="character"),
  function (object, value, ...)
    if (has(object, value))
      return(freqTable(object)[[value]])
    else
      return (0))

# Returns the document-token frequency matrix of this view, optionally
# weighted.
setMethod("documentTokenMatrix", 
          signature(object = "TokenizedCorpusView", 
                    weight = "missing"),
  function(object, weight=weightTf)
    if (is(ref(object, "documentTokenMatrix"), "NULLTermDocMatrix"))
      (ref(object, "documentTokenMatrix") <- 
        makeDocumentTokenMatrix(lapply(documentViews(object), freqTable),
                                unique(object),
                                names(documentViews(object))))
    else
      ref(object, "documentTokenMatrix"))

setMethod("documentTokenMatrix", 
          signature(object = "TokenizedCorpusView", 
                    weight = "WeightFunction"),
  function(object, weight=weightTf) 
    weightTDM(documentTokenMatrix(object), weight))

# UTILITES

# prints out the corpus id, total number of tokens, and total number of 
# unique tokens
setMethod("show", "TokenizedCorpusView",
  function(object) {
    nUnique <- length(unique(object))
    nTotal <- sum(documentTokenMatrix(object)@Data)

    str <- paste("A tokenized corpus view with", nTotal, 
                 "total tokens and", nUnique, "unique tokens\n")
    cat(str)
  }
)
